import { HASHTAG_FILTER, MASTODON_INSTANCE, ODESLI_API_KEY } from '$env/static/private';
import { log } from '$lib/log';
import type { Post, Tag, TimelineEvent } from '$lib/mastodon/response';
import type { OdesliResponse, Platform, SongInfo } from '$lib/odesliResponse';
import { getPosts, savePost } from '$lib/server/db';
import { createFeed, saveAtomFeed } from '$lib/server/rss';
import { sleep } from '$lib/sleep';
import { WebSocket } from 'ws';

const URL_REGEX = new RegExp(/href="(?<postUrl>[^>]+?)" target="_blank"/gm);

export class TimelineReader {
  private static _instance: TimelineReader;

  public static async getSongInfoInPost(post: Post): Promise<SongInfo[]> {
    const urlMatches = post.content.matchAll(URL_REGEX);
    const songs: SongInfo[] = [];
    for (const match of urlMatches) {
      if (match === undefined || match.groups === undefined) {
        log.warn('Match listed in allMatches, but either it or its groups are undefined', match);
        continue;
      }
      const urlMatch = match.groups.postUrl.toString();
      let url: URL;
      try {
        url = new URL(urlMatch);
      } catch (e) {
        log.error('URL found via Regex does not seem to be a valud url', urlMatch, e);
        continue;
      }

      // Check *all* found url and let odesli determine if it is music or not
      log.debug(`Checking ${url} if it contains song data`);
      const info = await TimelineReader.getSongInfo(url);
      log.debug(`Found song info for ${url}?`, info);
      if (info) {
        songs.push(info);
      }
    }
    return songs;
  }

  private static async getSongInfo(url: URL, remainingTries = 6): Promise<SongInfo | null> {
    if (remainingTries === 0) {
      log.error('No tries remaining. Lookup failed!');
      return null;
    }
    if (url.hostname === 'songwhip.com') {
      // song.link doesn't support songwhip links and songwhip themselves will provide metadata if you pass in a
      // Apple Music/Spotify/etc link, but won't when provided with their own link, so no way to extract song info
      // except maybe scraping their HTML
      return null;
    }

    const odesliParams = new URLSearchParams();
    odesliParams.append('url', url.toString());
    odesliParams.append('userCountry', 'DE');
    odesliParams.append('songIfSingle', 'true');
    if (ODESLI_API_KEY && ODESLI_API_KEY !== 'CHANGE_ME') {
      odesliParams.append('key', ODESLI_API_KEY);
    }
    const odesliApiUrl = `https://api.song.link/v1-alpha.1/links?${odesliParams}`;
    try {
      return fetch(odesliApiUrl).then(async (response) => {
        if (response.status === 429) {
          throw new Error('Rate limit reached', { cause: 429 });
        }
        const odesliInfo: OdesliResponse = await response.json();
        if (!odesliInfo || !odesliInfo.entitiesByUniqueId || !odesliInfo.entityUniqueId) {
          return null;
        }
        const info = odesliInfo.entitiesByUniqueId[odesliInfo.entityUniqueId];
        const platform: Platform = 'youtube';
        return {
          ...info,
          pageUrl: odesliInfo.pageUrl,
          youtubeUrl: odesliInfo.linksByPlatform[platform]?.url,
          postedUrl: url.toString()
        } as SongInfo;
      });
    } catch (e) {
      if (e instanceof Error && e.cause === 429) {
        log.warn('song.link rate limit reached. Trying again in 10 seconds');
        await sleep(10_000);
        return await this.getSongInfo(url, remainingTries - 1);
      }
      log.error(`Failed to load ${url} info from song.link`, e);
      return null;
    }
  }

  private startWebsocket() {
    const socket = new WebSocket(`wss://${MASTODON_INSTANCE}/api/v1/streaming`);
    socket.onopen = () => {
      log.log('Connected to WS');
      socket.send('{ "type": "subscribe", "stream": "public:local"}');
    };
    socket.onmessage = async (event) => {
      try {
        const data: TimelineEvent = JSON.parse(event.data.toString());
        if (data.event !== 'update') {
          return;
        }
        const post: Post = JSON.parse(data.payload);
        const hashttags: string[] = HASHTAG_FILTER.split(',');
        const found_tags: Tag[] = post.tags.filter((t: Tag) => hashttags.includes(t.name));

        const songs = await TimelineReader.getSongInfoInPost(post);

        // If we don't have any tags or non-youtube urls, check youtube
        // YT is handled separately, because it requires an API call and therefore is slower
        if (songs.length === 0 && found_tags.length === 0) {
          log.log('Ignoring post', post.url);
          return;
        }

        await savePost(post, songs);
        log.debug('Saved post', post.url);

        const posts = await getPosts(null, null, 100);
        await saveAtomFeed(createFeed(posts));
      } catch (e) {
        log.error('error message', event, event.data, e);
      }
    };
    socket.onclose = (event) => {
      log.warn(
        `Websocket connection to ${MASTODON_INSTANCE} closed. Code: ${event.code}, reason: '${event.reason}'`
      );
      setTimeout(() => {
        log.info(`Attempting to reconenct to WS`);
        this.startWebsocket();
      }, 10000);
    };
    socket.onerror = (event) => {
      log.error(
        `Websocket connection to ${MASTODON_INSTANCE} failed. ${event.type}: ${event.error}, message: '${event.message}'`
      );
    };
  }

  private constructor() {
    this.startWebsocket();
  }

  public static init() {
    if (this._instance === undefined) {
      this._instance = new TimelineReader();
    }
  }

  public static get instance(): TimelineReader {
    TimelineReader.init();
    return this._instance;
  }
}
